{-----------------------------------------------------------------------}
{  DC-Stegano v1.0 by Nikola Injac in 1997.  Hiding Data in Pic-LSBits  }
{  email: Nikola_Injac@hotmail.com                                      }
{  URL: http://members.tripod.com/~Nikola_Injac/stegano/                }
{-----------------------------------------------------------------------}
{  Syntax: DCstego -a Carrier.PCX SecretFile OutSteg.PCX                }
{          DCstego -e OutSteg.PCX                                       }
{          DCstego -i Picture.PCX                                       }
{-----------------------------------------------------------------------}
{  Stego Header: 0123456789ABCDE ... 1F3F                               }
{                LLNNNNNNNNEEEE[data....]                               }
{                                                                       }
{  LL ............ FileLength (Word)                                    }
{  NNNNNNNN ...... FileName (8 Bytes)                                   }
{  EEEE .......... FileExtension (1+3 Bytes)                            }
{  [data] ........ Data (7986 Bytes, best case)                         }
{-----------------------------------------------------------------------}
{ 1. After the carrier pic loaded, the palette is scanned for similar   }
{    colors - those wich substitution is hard to notice with bare eye.  }
{ 2. The carrier maximum capacity is calculated (UsableTable).          }
{ 3. A portion of the Secret file (max. capacity) is loaded into buffer.}
{ 4. Length and filename are added to the header.                       }
{ 5. Buffer (Data) is coded in the suitable carrier LSBs.               }
{ 6. Stegano Pic is saved.  / DeSteg is vice versa story.               }
{-----------------------------------------------------------------------}

Program DC_Stegano;

Uses DOS,
     NicePCX,                             {Load/Save PCX in 320x200/256}
     Logo;       {Only to write the ending logo-sequence, not important}

CONST ToleranceSingle  = 5;  {You can vary this values in order to gain}
      ToleranceSumm    = 13;                     { more color accuracy.}

TYPE TMsgBuffer  = ARRAY[0..8000] OF BYTE;     {Max. hidable Bytes=8000}
     TUsableTable= ARRAY[0..127] OF BOOLEAN;  {}

VAR
     Capacity    : WORD;
     Lengt       : WORD;
     RealFileLen : WORD;
     MsgBuffer   : TMsgBuffer;
     UsableTable : TUsableTable;
     Command     : String;
     Written     : String;

{-----------------------------------------------------------------------}
{- Writes the Logo Information
{-----------------------------------------------------------------------}
PROCEDURE WriteLogo;
BEGIN
  ByeBye;
  Write(#10#10#10#10#10#10#10#10#10);
END;

{-----------------------------------------------------------------------}
{- Secret fileName is coded along with its data;
{- Form: NNNNNNNN.EEE -> always 12 bytes used
{-----------------------------------------------------------------------}
FUNCTION GetOnlyFilename(Path:String):String;
VAR Dir    : DirStr;
    Nam    : NameStr;
    Ext    : ExtStr;
    MyName : String;

BEGIN
  FSplit(Path, Dir, Nam, Ext);
  MyName:=Nam+Ext;
  WHILE (Length(MyName)<12) DO MyName:=MyName+' ';
  GetOnlyFilename:=MyName;
END;

{-----------------------------------------------------------------------}
{- Look for similar pair and unpair colors (depending on Tolerance)
{-----------------------------------------------------------------------}
PROCEDURE MakeUsableColTable(Pal:TPal; VAR Table:TUsableTable);
VAR i,j, Clr: WORD;
BEGIN
  FOR i:=0 TO 127 DO BEGIN
    Table[i]:=TRUE;
    Clr:=0;
    FOR j:=0 TO 2 DO
      IF ABS(Pal[i*6+j]-Pal[i*6+j+3])>ToleranceSingle THEN
        Table[i]:=FALSE
      ELSE Inc(Clr,(ABS(Pal[i*3+j]-Pal[i*3+j+3])));
    IF Table[i] THEN IF Clr>ToleranceSumm THEN Table[i]:=FALSE;
  END;
END;

{-----------------------------------------------------------------------}
{- Count the max. Capacity a picture can take (name & len excluded)
{-----------------------------------------------------------------------}
FUNCTION GetCapacity(Table:TUsableTable):WORD;
VAR i, TempCap : WORD;
BEGIN
  TempCap:=0;
  FOR i:=0 TO 63999 DO
    IF Table[Mem[$a000:i] SHR 1] THEN Inc(TempCap);
  TempCap:=TempCap SHR 3;
  IF TempCap<16 THEN
    Fault('Unoptimized picture palette - low capacity.')
  ELSE GetCapacity:=TempCap-2-12;
END;

{-----------------------------------------------------------------------}
{- Attemp to load the entire Secret file
{-----------------------------------------------------------------------}
PROCEDURE LoadSecretFile(FName:String);
VAR
  F           : FILE;
  i           : WORD;
  OnlyFileName: String;
BEGIN
  Assign(F,FName);
  {$I-} Reset(F,1); {$I+}
  IF IOResult<>0 THEN Fault('"'+FName+'" could not be opened.');
  RealFileLen:=FileSize(F);
  BlockRead(F,MsgBuffer[14], Capacity, Lengt);
  MsgBuffer[0]:=Lo(Lengt);
  MsgBuffer[1]:=Hi(Lengt);
  OnlyFileName:=GetOnlyFilename(FName);
  FOR i:=1 TO 12 DO MsgBuffer[1+i]:=Ord(OnlyFilename[i]);
  Close(F);
END;

{-----------------------------------------------------------------------}
{- Encode - hide Data!
{-----------------------------------------------------------------------}
PROCEDURE StegEmUp;
VAR
   i, Shifter  : WORD;
   ShifterExit : WORD;
   MsgByte     : BYTE;
BEGIN
   i:=0;
   ShifterExit:=(Lengt+2+12) SHL 3; {Header 14Bytes}
   Shifter:=0;
   REPEAT
     MsgByte:=Mem[$a000:i]; {Get Byte}
     IF UsableTable[MsgByte SHR 1] THEN BEGIN
       Mem[$a000:i]:=(MsgByte AND 254) OR
           ((MsgBuffer[Shifter SHR 3] SHR (7-Shifter mod 8)) AND 1);
       INC(Shifter); {Next Text-Bit}
     END;
     INC(i); {Next Pic-Byte}
   UNTIL Shifter=ShifterExit;
END;

{-----------------------------------------------------------------------}
{- Decode - seek Data!
{-----------------------------------------------------------------------}
PROCEDURE DeSteg;
VAR i,WrOffs   : WORD;
    Rotate     : BYTE;
    SrcByte    : BYTE;
    WrByte     : BYTE;
BEGIN
  WrOffs:=0;
  WrByte:=0;
  Rotate:=0; {Bit Shift}
  FOR i:=0 TO 63999 DO BEGIN
    SrcByte:=Mem[$a000:i]; {Get Pic Byte}
    IF UsableTable[SrcByte SHR 1] THEN BEGIN
      WrByte:=(WrByte SHL 1) OR (SrcByte AND 1);
      Inc(Rotate);
      IF Rotate=8 THEN BEGIN
        Rotate:=0;
        MsgBuffer[WrOffs]:=WrByte;
        INC(WrOffs); {Next Write-Byte}
      END;
    END;
  END;
END;

{-----------------------------------------------------------------------}
{- Save Out-secret file
{-----------------------------------------------------------------------}
PROCEDURE SaveOutFile(VAR Written:String);
VAR  F        : FILE;
     OutName  : String[12];
     i        : Word;
BEGIN
  OutName:='';
  FOR i:=2 TO 13 DO
    IF (MsgBuffer[i]<>32) THEN OutName:=OutName+Chr(MsgBuffer[i]);
  IF OutName='' THEN OutName:='Unknown';
  Assign(F,OutName);
  {$I-} Rewrite(F,1); {$I+}
  IF IOResult<>0 THEN Fault('Could not write "'+OutName+'".');
  BlockWrite(F,MsgBuffer[14],Lengt,i);
  Close(F);
  Written:=OutName;
END;

{-----------------------------------------------------------------------}
Procedure WriteHelp;
BEGIN
  WriteLogo;
  WriteLn('Stego Syntax:'#10);
  WriteLn('Encode: DCstego.exe -e Carrier.PCX SecretFile Stego.PCX');
  WriteLn('Decode: DCstego.exe -d Stego.PCX');
  WriteLn('Info  : DCstego.exe -i Picture.PCX'#10);
  Halt;
END;

{-----------------------------------------------------------------------}
Procedure CheckParameters(Cnt:WORD);
VAR Comman : String;
    i      : Word;
BEGIN
  IF ParamCount<>Cnt THEN BEGIN
    WriteLn('Missing Parameters!'#10);
  Halt(0);
  END;
END;

{***********************************************************************}
{*                             --- MAIN ---
{***********************************************************************}
BEGIN
  IF ParamCount=0 THEN WriteHelp;
  Command:=ParamStr(1);
  CASE UpCase(Command[2]) OF
  'E','S': BEGIN  {ENCODING----------------------------------------------}
          CheckParameters(4);
          LoadPCX(ParamStr(2));
          MakeUsableColTable(Pal, UsableTable);
          Capacity:=GetCapacity(UsableTable);
          LoadSecretFile(ParamStr(3));
          StegEmUp;
          SavePCX(ParamStr(4));
          ASM xor ax,ax;int 16h; END; {Pause}
          WriteLogo;
          Write('Secret File ("',ParamStr(2),'", ',RealFileLen,' bytes) ');
          IF RealFileLen>Capacity THEN
            WriteLn('too long.'#10#13'Only ',Capacity,' bytes written! Try to reduce.')
            ELSE
            WriteLn('hidden successfully.');
        END;
  'D': BEGIN  {DECODING-------------------------------------------------}
          CheckParameters(2);
          LoadPCX(ParamStr(2));
          MakeUsableColTable(Pal, UsableTable);
          Capacity:=GetCapacity(UsableTable);
          DeSteg;
          Lengt:=MsgBuffer[1]*256+MsgBuffer[0];
          IF Lengt>Capacity THEN Fault('No secret data hidden.');
          SaveOutFile(Written);
          WriteLogo;
          Write('Secret File "',Written,'" written.');
       END;
  'I': BEGIN  {Info-----------------------------------------------------}
          CheckParameters(2);
          LoadPCX(ParamStr(2));
          MakeUsableColTable(Pal, UsableTable);
          Capacity:=GetCapacity(UsableTable);
          WriteLogo;
          Write('Capacity of "',ParamStr(2),'" is ',Capacity,' Bytes.');
       END;
  ELSE
    WriteHelp;
  END;
END.
{**************************** THE END **********************************}
